#include <cstring>
#include <cstdio>
#include <cstdlib>

#include "configg.h"
#include "basstr.h"
#include "AuxFuncs.h"
#include "AuxRND.h"
#include "phong.h"

// This implements methods of a general interface

// -----------------------------------------------------------------------------------
// Implementation using Phong model - white color only

BRDF_Phong::BRDF_Phong(float rho_diffuse, float rho_specular, float spec_indexV)
{
  rho_diff = rho_diffuse;
  rho_spec = rho_specular;
  spec_index = spec_indexV;
}

void
BRDF_Phong::getString(string &outstr) const
{
  char str[200];
  sprintf(str, "Phong rhod=%3.2f rhos=%3.2f si=%3.3f",
	  rho_diff, rho_spec, spec_index);
  outstr.assign(str);
}

// returns cosine corrected value of BRDF given the illumination
// and viewer angle in RGB, returns false on success
bool
BRDF_Phong::getBRDFValue(int irow, int jcol, float theta_v,
			 float phi_v, float theta_i, float phi_i,
			 float RGB[3]) const
{
  //cout << "A" << std::flush;
  // Phi-angles
  // illumination is this way reflected different way
  const float rp_i = (phi_i + M_PI);
  const float rp_v = phi_v;
  // Theta-angles
  const float rt_i = theta_i;
  const float rt_v = theta_v;

  // The vectors of illumination and view vector
  float refl_illum_vec[3], view_vec[3];
  ConvertThetaPhiToXYZ2(rt_i, rp_i, refl_illum_vec); // ideally reflected ray
  ConvertThetaPhiToXYZ2(rt_v, rp_v, view_vec); // viewing vector
  // cosine angle between reflected ray and viewing direction
  float cosangle = refl_illum_vec[0] * view_vec[0] +
    refl_illum_vec[1] * view_vec[1] +
    refl_illum_vec[2] * view_vec[2];
  if (cosangle < 0.f) cosangle = 0.f;
  float pexp = powf(cosangle, fabs(spec_index));
  float brdf = rho_diff/M_PI + rho_spec * (spec_index + 2.f)/(2.f*M_PI) * pexp;
  assert(!isnan(brdf));
  // the clamp of negative values due to the specular lobe
  if (brdf < 1e-25)
    brdf = 1e-25;
  // final value is cosine corrected for theta_i
  RGB[0] = RGB[1] = RGB[2] = brdf * cos(rt_i);
  return false; // OK
}

bool 
BRDF_Phong::spawnStochasticRay(int irow, int jcol, // position in the texture
			       float rnd1, float rnd2, // input RND numbers
			       float theta_view, float phi_view, // the direction of a viewer
			       float XYZ[3], // output direction
			       float RGB[3], // corresponding color
			       float &prob) const // and probability
{
#ifdef TEST_STOCHASTIC
  // this is testing a general interface against phong model - only test!!!
  return BRDF::spawnStochasticRay(irow, jcol, rnd1, rnd2,
				  theta_view, phi_view, XYZ, RGB, prob);  
#endif
  irow = irow;
  jcol = jcol;
  // z-axis is the normal of a surface (0,0,1)
  float rho_total = rho_diff + rho_spec;
  float valueEval = 0.f;
  // value to be used
  //  float value = 2.f*rho_diff+(spec_index+2.f)/(spec_index+1.f)*rho_spec;
  if (rnd1 < rho_diff/rho_total) {
    // sample diffuse par as rho_d*cos(theta)
    rnd1 = rnd1 / rho_diff * rho_total; // normalize the random variable to <0,1>
    assert(rnd1 <= 1.000001f);
    assert(0.f <= rnd1);
    float phi = 2.0f*M_PI*rnd1; // tmp2
    float sinTheta = sqrt(1.0f - rnd2); // sin(theta)
    XYZ[0] = cos(phi) * sinTheta;
    XYZ[1] = sin(phi) * sinTheta;
    XYZ[2] = sqrt(rnd2); // cos(theta)
    prob = XYZ[2] / M_PI; // prob given by PDF is cos(theta)/M_PI
    // compute reflected ray
    //float R[3];
    //ConvertThetaPhiToXYZ2(theta_view, phi_view + M_PI, R);
    //float cosAlpha = XYZ[0]*R[0]+XYZ[1]*R[1]+XYZ[2]*R[2];    
    //prob = XYZ[2] / M_PI; // prob given by PDF is cos(theta)/M_PI
    // where BRDF is only for the diffuse part
    RGB[0] = RGB[1] = RGB[2] = XYZ[2]/M_PI * rho_total; 
    // Note - multiplication by rho_total above is necessary, as in fact
    // we compute cos(theta)*rho_d/M_PI *(rho_spec+rho_d)/rho_d
    // The results were werified numerically
    return false; // OK
  }

  // sample specular lobe in local coordinate system x,y,z  
  if (rho_spec < 1e-10f) {
    RGB[0] = RGB[1] = RGB[2] = 0.0f;
    prob = 1.f / M_PI; // prob given by PDF of diffuse component
    return true; // no ray does exist here
  }
  // alpha ... angle between perfect mirror reflection and the outgoing direction
  rnd1 = (rnd1*rho_total - rho_diff)/rho_spec; // normalize the random variable to <0,1>
  //cout << rnd1 << " ";
  assert(rnd1 >= 0.f);
  assert(rnd1 <= 1.00001f);
  // compute the angle alpha, it is enough its cos(alpha) and sin(alpha)
  float cosAlpha = powf(rnd1, 1.0f/(spec_index+1.0f)); // our sampling distribution
  //float aa = 180.f/M_PI*acos(cosAlpha);
  //if (aa > 60.f) cout << aa << " ";
  assert(cosAlpha <= 1.0f);
  assert(cosAlpha >= 0.f);
  float sinAlpha = sqrt(1.0f - cosAlpha*cosAlpha);
  float XYZL[3]; // the result in local coordinate system given by reflected ray
  float phi = 2.0f*M_PI*rnd2;
  // we compute the direction in coordinate system (1,0,0)x(0,1,0)x(0,0,1)
  XYZL[0] = cos(phi) * sinAlpha;
  XYZL[1] = sin(phi) * sinAlpha;
  XYZL[2] = cosAlpha;
  // we transform now the vector to the coordinate system given
  // by ideally reflected ray R,U,V
  float R[3], U[3], V[3];
  ConvertThetaPhiToXYZ2(theta_view, phi_view + M_PI, R);
  ArbitraryNormal(R, U); // generate normal to R  
  vec_crossprod(R, U, V); // and the V to create the coordinate system R,U,V
  
  // transform XYZL in the coordinate system R,U,V to X,Y,Z coordinate system
  // first compute Z-value
  XYZ[2] = XYZL[0] * U[2] + XYZL[1] * V[2] + XYZL[2] * R[2];
  // now we check, if the samples goes below the surface
  if ((XYZ[2] <= 1e-10f)||(cosAlpha<1e-10)) {
    // generated direction is below the surface specified by normal
    RGB[0] = RGB[1] = RGB[2] = 0.0f; // black colour
    XYZ[0] = 0; XYZ[0] = 1; XYZ[2] = 0;
    prob = 1.0f;
    return true; // no ray can be shot for given two random values
  }
  // Then compute two other coordinates X an Y
  XYZ[0] = XYZL[0] * U[0] + XYZL[1] * V[0] + XYZL[2] * R[0];
  XYZ[1] = XYZL[0] * U[1] + XYZL[1] * V[1] + XYZL[2] * R[1];
  assert(cosAlpha > 0.f);
  // Let us evaluate the probability of sampling of the specular lobe
  float cosAlphaPoweredFactor = pow(cosAlpha, spec_index)/(2.f*M_PI);
  prob = (spec_index+1.f)*cosAlphaPoweredFactor;
  assert(prob > 0.f);
  // This formula contains cosine(theta_illum) in the result by XYZ[2]
  // In fact it is reevaluating BRDF for the lighing and viewer direction
  assert(XYZL[2] >= 0.f);
  float brdfpart = powf(XYZL[2], spec_index)/(2.f*M_PI);
  float brdf = (spec_index+2.f) * brdfpart * rho_total;
  RGB[0] = RGB[1] = RGB[2] = brdf * XYZ[2]; // return brdf*cos(theta_il)
  // Note - multiplication by rho_total above is necessary, as in fact we
  // compute cos(theta)*rho_spec*(spec_index+2)/2/M_PI *cos^spec_index(alpha)*(rho_spec+rho_d)/rho_d
  // The results were werified numerically on various ranges of rho_spec, rho_diff, spec_index
  // Vlastimil Havran, 2013/09/10
  return false; // OK
}

// returns probability of sampling in a given viewer direction
// The RGB value has to be provided from fomer call of function getBRDFValue.
// It shall not be used in combination of importanceSampling - it returns different
// value - since importance sampling is probabilistic and here we compute some average.
// It shall be used only if we sample from another distribution and aske what it is
// probabibility of chosing this direction, such as in MIS.
float
BRDF_Phong::getProbOfSampling(int irow, int jcol,
			      float theta_view, float phi_view,
			      float theta_i, float phi_i,
			      const float RGB[3]) const
{
  // general algorithm, which works correctly
  // value does exist as is provided by user
  float albedo = getAlbedo(irow, jcol, theta_view, phi_view);
  float prob = luminance(RGB)/albedo + 1e-25; // add small value to avoid zero probability
  return prob;
}

float
BRDF_Phong::getAlbedo(int irow, int jcol, // texture position
		      // viewer direction as two angles in radians
		      float theta_view, float phi_view) const
{ 
#ifdef TEST_STOCHASTIC
  // testing albedo computation here
  float albedo = BRDF::getAlbedo(irow, jcol, theta_view, phi_view);
  //  cout << " al= " << albedo;
  return albedo;
#else
  // analytic formula
  return rho_diff + rho_spec;
#endif  
}

